#!/bin/bash
########################################################################
# build_gcc.sh
#
# Build GCC for UPS.
#
########################################################################

####################################
# Useful variables
prog=${BASH_SOURCE[0]##*/}
thisdir=$(cd "$(dirname "${BASH_SOURCE[0]}")" && pwd -P)
trace=maybe_xtrace # Could set to xtrace instead for unconditional.

####################################
# Useful functions.

# Report usage.
usage() {
  cat <<EOF
USAGE: ${prog} ${ssi_msg_config_usage:-<ssi-msg-config-options>} [--] <product-dir> [tar]
       ${prog} -h|--help [<product-dir>]

EOF
  (( ${1:-0} < 2 )) && return;
  cat <<EOF
OPTIONS

  -h|--help

     This help.

${ssi_msg_config_options:-"  <ssi-msg-config-options>

    [Supply <product-dir> to enable more details]
"}\
ARGUMENTS

  <product-dir>

    The top-level UPS directory in which the source code distribution
    should be constructed.

  [tar]

    Generate a binary installation archive.

ENVIRONMENT

  DEBUG_BUILD
    Full command tracing for the build (VERBOSE=1)

  DEBUG_PATCHES
  VERBOSE_PATCHES
    Command tracing and increased verbosity for source unpacking and patching
    operations. DEBUG_PATCHES will cause exit after patches have been applied.

${ssi_msg_config_environment}
EOF
}

# Local (dumb) ssi_die() function for use until we source ssi_functions.
ssi_die() {
  printf "FATAL ERROR: ${*}\n" 1>&2; exit 1
}

# Make sure we have UPS and ssibuildshims set up.
ensure_ssibuildshims() {
  if ! (( $# )) || [ "${@: -1}" = "-h" ] || [ "${@: -1}" = "--help" ]; then
    # We\'ve been asked for help: give as much as we can.
    usage 2; exit 1
  fi
  # Find <product-dir> from our arguments:
  local arg
  for arg in "${@}"; do
    [[ "$arg" ==  -* ]] && continue;
    product_dir="${arg}"; break
  done
  if [ -z "${product_dir}" ]; then
    # Bad usage.
    ssi_die "Bad <product-dir> ${product_dir}\n$(usage)"
    exit 2
  elif ! { [ -d "${product_dir}/.upsfiles" ] && [ -w "${product_dir}" ]; }; then
    # Something wrong with <product-dir>.
    ssi_die "<product-dir> not a valid, writable unified UPS directory: ${product_dir}"
  fi
  # Promising...
  local abs_product_dir="$(cd "${product_dir}" 2>/dev/null 2>&1 && pwd -P)"
  (( $? )) && ssi_die "unable to cd to <product-dir>: ${product_dir}"
  # Absolute is safer.
  product_dir="${abs_product_dir}"
  # Find our own bootstraps and yank 'em.
  local ssibuildshims_topdir="${product_dir}/ssibuildshims/${ssibuildshims_version}"
  [ -f "${ssibuildshims_topdir}/bin/ssi_functions" ] && \
    source "${ssibuildshims_topdir}/bin/ssi_functions" >/dev/null 2>&1 && \
    ensure_ups "${product_dir}" && \
    setup ssibuildshims ${ssibuildshims_version} -z "${product_dir}" || \
    ssi_die "unable to find ssibuildshims ${ssibuildshims_version} in
${product_dir} and set it up with UPS."
}

# Verify that GCC has been built successfully.
gcc_built() {
  verify_in_filesystem "${@}" "${pkgdir}/lib"{64,}"/libstdc++.${shext}" && \
    verify_in_filesystem "${@}" "${pkgdir}/lib"{64,}"/libgfortran.${shext}" && \
    verify_in_filesystem "${@}" -x "${pkgdir}/bin/gcc" && \
    verify_in_filesystem "${@}" -x "${pkgdir}/bin/g++" && \
    verify_in_filesystem "${@}" -x "${pkgdir}/bin/gfortran" && \
    verify_in_filesystem "${@}" -d "${pkgdir}/include/c++"
} 2>&1

# Make the distribution and exit if successful.
maybe_make_distribution() {
  if [ "${maketar}" = "tar" ]; then
    gcc_built "$@" && make_basic_distribution && exit
  fi
}

# Expand and patch sources.
unpack_and_patch() {
  # Failed globs should be empty: save the shell state and set it to
  # what we want.
  ssi_autosave_shell_state; shopt -s nullglob

  local source_stamp="${sourcedir}/sources.stamp"
  local patch_stamp="${sourcedir}/patches.stamp"
  local pkgs=(gcc) pkgtar pkg

  # Unwind the GCC source first.
  local gcctar="${tardir}/${pkgtardir}.tar.gz"

  ${trace} prepare_source "${ssi_args[@]}" \
    -d "${sourcedir}" "${gcctar}" || exit $?

  # Now unwind all the dependency sources---no patching yet!
  for pkgtar in "${tardir}"/*.tar*; do
    [ "${pkgtar}" == "${gcctar}" ] && continue;
    pkg=${pkgtar##*/}
    pkg=${pkg%%-*}
    pkgs+=("${pkg}")
    ${trace} prepare_source "${ssi_args[@]}" \
      -d "${gcc_topdir}/${pkg}" \
      -e "--strip-components,1" "${pkgtar}" || exit $?
  done

  # Remember the current time.
  touch "$source_stamp"

  # Now check to see if we have a patch rollup archive newer than any of our
  # patches that we can use to bypass the patch application and the possible
  # need for regeneration of configure scripts. This became necessary on
  # 2020-03-30 when the mpfr-4.0.2 rollup patch as dynamically downloaded by
  # bootstrap.sh was updated to touch files triggering a regeneration of the
  # configure scripts. We must be very careful with dates to avoid re-triggering
  # the regeneration step on a machine that might not have an autotools suite of
  # sufficiently recent vintage.
  if [ -f "${rollup}" ] && \
       (( $(find "${patchdir}" -name '*.patch' -newer "${rollup}" \
                 2>/dev/null | wc -l) == 0 )); then
    ${trace} prepare_source "${ssi_args[@]}" \
      -d "${sourcedir}" "${rollup}" || \
      exit $?
    touch "${patch_stamp}"
    return
  fi

  # OK, we need to apply the patches and then the patch rollup archive must be
  # generated out-of-band using the ${patchdir}/make-patch-rollup script
  # according to ${patchdir}/00README.txt.
  local pad
  for pkg in "${pkgs[@]}"; do
    pad=${pkg#gcc}
    ${trace} patch_source "${ssi_args[@]}" \
      -a "${gcc_topdir}/${pad}" \
      "${patchdir}/${pkg}-"*.patch || \
      exit $?
  done
  touch "${patch_stamp}"

  # Stop with error for out-of-band rollup archive generation.
  ssi_die -i "patch rollup archive ${rollup} must be (re-)generated.\n" \
        "See patch/00README.txt"
}

########################################################################
# Main.
########################################################################

####################################
# Package name and version information.
package=gcc
origpkgver=v9_3_0
pkgver=${origpkgver}
ssibuildshims_version=v2_00_00
origpkgdotver=${origpkgver//_/.}; origpkgdotver=${origpkgdotver#v}
pkgdotver=${pkgver//_/.}; pkgdotver=${pkgdotver#v}
pkgtardir=gcc-${origpkgdotver}

####################################
# Handle options and arguments.

# Set up ssibuildshims and UPS / product_dir as early as possible.
ensure_ssibuildshims "${@}"

# Initialization and defaults.
(( ssi_trace = 1 ))

# Parse options.
while true; do
  case $1 in
    -h|--help) usage 2; exit 1;;
    --) shift; break;;
    -*) ssi_handle_config_options "${1}" || \
      ssi_die 2 "unrecognized option $1\n$(usage)";;
    *) shift; break # Swallow <product-dir>: dealt with.
  esac
  shift
done

# Environment overrides.
ssi_honor_environment_config

##################
# Remaining arguments.
maketar=${1}; shift

# Define globals.
define_basics NO_QUAL NO_DEBUG_SOURCE || exit

# Possible short-circuit.
maybe_make_distribution -q

echo "Building ${package} for ${OS}-${plat} (flavor ${flvr})"

# Top level of GCC source.
gcc_topdir="${sourcedir}/${pkgtardir}"

# Generated patch rollup archive (may include regenerated configure files).
rollup="${patchdir}/patch-rollup-${pkgdotver}.tar.bz2"

####################################
# Ensure directories.
mkdir -p "${pkgdir}" || \
  ssi_die "failed to create installation directory ${pkgdir}"
# Remove any existing sourcedir to avoid patch conflicts.
rm -rf "${sourcedir}" && mkdir -p "${sourcedir}" || \
  ssi_die "failed to create source directory ${sourcedir}"
# Clean build.
rm -rf "${blddir}" && mkdir -p "${blddir}" || \
    ssi_die "failed to create build directory ${blddir}"
####################################

# Expand and patch sources.
unpack_and_patch

# Exit if debugging.
[ -n "${DEBUG_PATCHES}" ] && exit 1

# Ascertain and configure parallelism.
num_cores=$(${SSIBUILDSHIMS_DIR}/bin/ncores)

####################################
# Attempt to sanitize the environment.

# Remove /usr/local/bin and . from path.
PATH=`dropit -e /usr/local/bin .`

# Remove /usr/local/X from LD_LIBRARY PATH and DYLD_LIBRARY_PATH
LD_LIBRARY_PATH=${LD_LIBRARY_PATH:+$(dropit -a -p "$LD_LIBRARY_PATH" /usr/local/)}
DYLD_LIBRARY_PATH=${DYLD_LIBRARY_PATH:+$(dropit -a -p "$DYLD_LIBRARY_PATH" /usr/local/)}

if [ ${OS1} = "Darwin" ]; then
  echo "DYLD_LIBRARY_PATH=$DYLD_LIBRARY_PATH"
  export DYLD_LIBRARY_PATH=${pkgdir}/lib64:${pkgdir}/lib
  shext=dylib
else
  echo "LD_LIBRARY_PATH=$LD_LIBRARY_PATH"
  export LD_LIBRARY_PATH=${pkgdir}/lib64:${pkgdir}/lib
  shext=so
fi

# Make sure libraries are in the path.
echo "COMPILER_PATH=$COMPILER_PATH"
export PATH=${pkgdir}/bin:${PATH}
export COMPILER_PATH=${pkgdir}/bin


echo ""
echo "LD_LIBRARY_PATH=$LD_LIBRARY_PATH"
echo "DYLD_LIBRARY_PATH=$DYLD_LIBRARY_PATH"
echo "COMPILER_PATH=$COMPILER_PATH"
####################################

cd "${gcc_topdir}" || \
  ssi_die "unable to cd to ${gcc_topdir} to build ${package}"

####################################
# Configure, build, and install.

##################
# Configure.

languages="c,c++,fortran,lto"
conf_options=(--prefix=${pkgdir} \
              --disable-libgcj \
              --enable-threads=posix \
              --enable-__cxa_atexit \
              --enable-languages=${languages} \
              --with-system-zlib \
              --enable-libstdcxx-time=rt \
              --enable-stage1-checking=yes \
              --enable-checking=release \
              --enable-plugin)
if [[ "$flvr" != "Linux64bit+2.6-2.12" ]]; then # not SLF6
  conf_options+=(--enable-lto)
  ######################################################################
  # Zstandard will be optional from GCC 10.1, leading to faster LTO
  # operations due to the significantly (8x) faster compression
  # performance.
  #    ( cd zstd && make -j ${num_cores} lib-release ) || \
  #        ssi_die "unable to build required zstd."
  #    conf_options+=("--with-zstd=$(pwd)/zstd")
  ######################################################################
fi
if [ "${OS1}" = Darwin ]; then
  sdk=$(xcrun --sdk macosx --show-sdk-path)
  [ -n "${sdk}" ] || ssi_die "unable to locate MacOS SDK"
  # Most headers are in the SDK now.
  conf_options+=(--with-native-system-header-dir=/usr/include \
                 --with-sysroot="${sdk}" )
  # Also, no support for 32-bit from Catalina on.
  [[ "${OS}" =~ ^d([0-9]+) ]] && (( ${BASH_REMATCH[1]} >= 18 )) && \
    conf_options+=(--disable-multilib)
else
  # Go not supported on Darwin targets.
  languages="${languages+${languages},}go"
fi

# Execute the configure script.
${trace} ${gcc_topdir}/configure "${conf_options[@]}" || \
  ssi_die "unable to configure ${package}"

##################
# Build.

# The profiledbootstrap target is reputed to produce a GCC which compiles code
# around 7% faster according to some benchmarks, but the build of GCC itself
# could take significantly longer on MacOS (>17h on a modern High Sierra
# laptop).
if [ -n "${NO_PROFILE_BOOTSTRAP}" ]; then
  target=profiledbootstrap
else
  target=bootstrap
fi

# Compose the build command.
build_cmd=(make ${target} -j "${num_cores}" ${DEBUG_BUILD:+VERBOSE=1})
[ "${OS1}" = Darwin ] && \
  build_cmd+=("BOOT_LDFLAGS=-Wl,-headerpad_max_install_names")

# Execute the build.
${trace} "${build_cmd[@]}" || ssi_die "unable to build ${package}"

##################
# Install.

${trace} make install || ssi_die "unable to install ${package}"

##################
# Verify install.

fail_msg=$(gcc_built)
(( $? == 0 )) || ssi_build_failed "${package}" "${fail_msg}"

####################################
# Installation post-processing.
# Remove .la files.
find "${pkgdir}" -name '*.la' | xargs rm -vf

# Make soft links for cc and f95. Note that c++ already exists.
cd "${pkgdir}/bin" || ssi_die "unable to cd to ${pkgdir}/bin"

[ -x cc ] || ln -s ${ssi_verbose:+-v} gcc cc
[ -x c++ ] || ln -s ${ssi_verbose:+-v} g++ c++
[ -x gfortran ] || ln -s ${ssi_verbose:+-v} gfortran f95

# Fix path references in Python GDB files
find "${pkgdir}/lib"* -name '*.py' | \
  xargs perl -wapi\~ -e \
    "s&(['\"])"'\Q'"${pkgdir}"'\E\b(.*?\g{1})&${1}\{prefix\}${2}.format(prefix=os.environ["GCC_FQ_DIR"])&g;'

# Declare the product to UPS.
${trace} declare_product "${ssi_args[@]}" \
         "${product_dir}" ${package} ${pkgver} NO_QUAL || exit $?

# Attempt to set up the package.
echo "Finished building ${package} ${pkgver}"

setup -B ${package} ${pkgver} -z "${product_dir}" -z "${PRODUCTS}" || \
  ssi_die "unable to set up newly-built product ${package} ${pkgver}"

echo "${package} is installed at ${pkgdir}"

# This must be last.
maybe_make_distribution
